/*  Basic solar system simulation with ViewGroups	*/
/*  implementing multiple views.			*/
/*  Uses Greenspan discrete mechanics as implented in	*/
/*  Grnspan.java					*/
/*  Provides GUI control of integration experiments.	*/
/*			Ernest Sibert - February 1997	*/

import java.awt.*;
import java.applet.*;
import java.awt.event.*;
import cpsJava.*;

public class NBApplt extends java.applet.Applet
		     implements Runnable, ActionListener
 {
   int  N;		/* Number of bodies	*/

   Body bodies[];	/* The bodies 		*/

   ViewGroup viewxyz;	/* Solar system - xy, zx planes	*/

   NBState state;	/* Current state, integration	*/

   Control ctrl;	/* Simulation controls		*/

   /*  Applet control items **********************************	*/

   Thread NBThread;		/* Runs the applet		*/

   Object safety;		/* For synchronization		*/

   int delay = 2000;		/* To slow it down (msec.)	*/

   boolean ready = false,	/* To synchronize startup	*/
	 ptrails = false,	/* Draw permanent trails	*/
	 running = false,	/* Is system running?		*/
	finished = false;	/* Is simulation finished?	*/

   String parm;			/* For applet parameters	*/

   int ns = 10,		/* Number of steps between plots	*/
       np = 10,		/* Number of times to plot		*/
       gs =  0,		/* Loop counter for plotting		*/
     ilim = 10;		/* Iteration limit			*/

   double eps = 1.0e-8,	/* Convergence criterion	*/
	   dt = 0.01,	/* Time step			*/
	width = 80.0;	/* Initil view width		*/


   /*  Initialize	*/

   public void init()
    {  safety = new Object();
       parm = getParameter("nbsys");
       if ( parm != null)
	{ Body wbodies[] = NBSystems.select(parm);
	  N = wbodies.length;
	  bodies = new Body[N];
	  for (int i = 0; i < N; i++) bodies[i] = wbodies[i];
	}
       System.out.println(parm+" "+N);
       setLayout(new FlowLayout(FlowLayout.CENTER));
       viewxyz = new ViewGroup();  add(viewxyz);
       viewxyz.init(2, 500, 500, 250, N);
       viewxyz.addView(1, 2,  3);	/* Standard x-y view	*/
       viewxyz.addView(1, 3, -2);	/* x-z viewed from -y	*/
       viewxyz.setView(width, 0.0, 0.0, 0.0);
       state = new Greenspan();  state.init(N, 0.0);
       installSystem();
       viewxyz.setControls();	/* Set view controls */
       parm = getParameter("vwidth");
       if ( parm != null ) width = Double.valueOf(parm).doubleValue();
       viewxyz.setView(width, 0.0, 0.0, 0.0);
       System.out.println(
	  "au->pix: "+viewxyz.scale+"  km->pix: "+viewxyz.bscale);
    /*  Integration parameters */
       parm = getParameter("deltat");
       if ( parm != null ) dt = Double.valueOf(parm).doubleValue();
       parm = getParameter("epsilon");
       if ( parm != null ) eps = Double.valueOf(parm).doubleValue();
       parm = getParameter("limit");
       if ( parm != null ) ilim = Integer.valueOf(parm).intValue();
       state.initInteg(dt, eps, ilim);
    /*  Integration time, plot controls */
       parm = getParameter("nsteps");
       if ( parm != null ) ns = Integer.valueOf(parm).intValue();
       parm = getParameter("nplots");
       if ( parm != null ) np = Integer.valueOf(parm).intValue();
       makeControl();		/* Experiment control panel */
       parm = getParameter("delay");
       if ( parm != null ) delay = Integer.valueOf(parm).intValue();
       parm = getParameter("ptrails");
       if ( parm != null ) ptrails = parm.equalsIgnoreCase("yes");
       viewxyz.setTrails(ptrails);
       ready = true;
    }

/* The control panel ==================================================	*/

   Parameter pDeltat, pIlim, pNsteps, 	// Numeric (gui) parameters
	     pNplots, pEps, cTime;	// and current time

   Button rerun,			// Triggers new run
          cntnu;			// Continues current run

   String ttl = "Integration controls";

   EasyFormat frmt = new EasyFormat();

   String ftim(double tm) { return frmt.fFormat(tm, 12, 5); }

   void makeControl()
    { ctrl = new Control();  ctrl.init(8, 1);  add(ctrl);
      ctrl.title(ttl);
      pDeltat = new Parameter();
      pDeltat.init("deltat", Double.toString(dt));
      ctrl.addItem(pDeltat);				// 1
      pNsteps = new Parameter();
      pNsteps.init("nsteps", Integer.toString(ns));
      ctrl.addItem(pNsteps);				// 2
      pNplots = new Parameter();
      pNplots.init("nplots", Integer.toString(np));
      ctrl.addItem(pNplots);				// 3
      pIlim = new Parameter();
      pIlim.init("ilim", Integer.toString(ilim));
      ctrl.addItem(pIlim);				// 4
      pEps = new Parameter();
      pEps.init("epsilon", Double.toString(eps));
      ctrl.addItem(pEps);				// 5
      rerun = new Button("Recalculate");
      ctrl.addItem(rerun);				// 6
      rerun.addActionListener(this);
      rerun.setActionCommand("R");
      cntnu = new Button("Continue");
      ctrl.addItem(cntnu);				// 7
      cntnu.addActionListener(this);
      cntnu.setActionCommand("C");
      cTime = new Parameter();
      cTime.init("time: ", ftim(state.stime));
      ctrl.addItem(cTime);
    }



  //  String -> float

  double toDouble(String s)
   { return Double.valueOf(s).floatValue(); }

  //  String -> int

  int  toInt(String s)
   { return Integer.parseInt(s); }


  //  Responding to the "Recalculate" or "Continue" buttons

  public void actionPerformed(ActionEvent e)
   { synchronized(safety)
      { dt   = toDouble(pDeltat.pValue());
        ns   = toInt(pNsteps.pValue());
        np   = toInt(pNplots.pValue());
        ilim = toInt(pIlim.pValue());
        eps  = toDouble(pEps.pValue());
        if (e.getActionCommand().charAt(0) == 'R')   /* Rerun */
	  { state.resetSystem(bodies);  state.stime = 0.0;
	    cTime.setValue(Double.toString(state.stime));
	  }
        state.initInteg(dt, eps, ilim);
	if ( ptrails )			/* Reinitialize drawing */
	 { viewxyz.setTrails(false);
	   viewxyz.drawAll(state.X0);
	   viewxyz.setTrails(true);
	 }
      }
     if ( running ) { run(); }
     else { start(); }
   }


/* Simulation management ==============================================	*/

   /*  Draw complete system (from scratch) using positions in Xs.	*/

   void drawSystem(double Xs[][])
    { synchronized (safety) { viewxyz.drawAll(Xs); }
    }


   /*  Install bodies in state, initialize G.				*/

   void installSystem()
    { NBObj ob;		/* Ignored for the moment	*/
      state.G = NBSystems.G;
      for ( int i = 0;  i < N;  i++ )
       { state.installBody( bodies[i] );
	 ob = viewxyz.loadObject(bodies[i]); 
       }
      
    }



   /*  Move bodies to current positions and replot - preliminary.	*/

   void moveSystem(double Xs[][])
    { viewxyz.drawAll(Xs);
    }


   /* Delay for specified time (in msec.) */

   void pause(int delay)
    { if ( delay > 0 )
       {  try { Thread.sleep(delay); }
          catch (InterruptedException ex) { }
       }
    }

   /* Start process	*/

   public void start()
    { while ( !ready  ) pause(200);		/* Wait for init */
      drawSystem(state.X0);
      if ( NBThread == null )
       { NBThread = new Thread(this);
	 running = true;
	 NBThread.start(); }
    }


   /* Stop process  */

   public void stop()
    { if ( NBThread != null )
       synchronized (safety)
	{ NBThread.stop();  running = false;  NBThread = null; }
    }



   /*  Run system ns*np steps, plotting every ns steps.	*/

   public void run()
    { String tmf;
      finished = false;
      viewxyz.repaint();
      if ( delay > 0) pause(delay);
      for ( gs = 1;  gs <= np;  )
       { int nac = advance(); drawSystem(state.X0);
	 if ( delay > 0 ) pause(delay);  viewxyz.repaint();
	 tmf = ftim(state.stime);
	 System.out.println("Time: "+ tmf
			   +"   Accelerations: "+nac
			   +"   Failures: "+state.fail);
	 cTime.setValue(tmf);
	 state.printState(state.X0, state.V0);
       }
      System.out.println("Finished.");
      finished = true;
    }


   /*  Advance over next group of time steps.  Returns total number	*/
   /*  of acceleration calculations (depends on integrator).		*/

   int advance()
    { int na;
      synchronized (safety) { na = state.nSteps(ns);  gs++; }
      return na;
    }

 }


